// Fill out your copyright notice in the Description page of Project Settings.

#include "LMWaveManager.h"

#include "Components/CapsuleComponent.h"
#include "Enemy/LMEnemy.h"

// Sets default values
ALMWaveManager::ALMWaveManager()
{
 	// Set this actor to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = true;

}

// Called when the game starts or when spawned
void ALMWaveManager::BeginPlay()
{
	Super::BeginPlay();
	StartWave();
}

void ALMWaveManager::SpawnEnemy(ALMSpawnPosition* spawnPosition)
{
	int RandomIndex = FMath::RandRange(0, AllEnemyType.Num()-1);
	while(EnemiesPerType[RandomIndex] <= 0 )
	{
		RandomIndex = FMath::RandRange(0, AllEnemyType.Num()-1);
	}

	//Choisis un ennemis à spawn random
	TSubclassOf<ALMEnemy> enemyToSpawn = AllEnemyType[RandomIndex];
	//Spawn sur une position
	FActorSpawnParameters SpawnParameters;
	SpawnParameters.SpawnCollisionHandlingOverride = ESpawnActorCollisionHandlingMethod::AdjustIfPossibleButAlwaysSpawn;
	ALMEnemy* tempEnemy = GetWorld()->SpawnActor<ALMEnemy>(enemyToSpawn, spawnPosition->GetActorLocation(), FRotator::ZeroRotator);
	
	if (!tempEnemy)
	{
		UE_LOG(LogTemp, Error, TEXT("Failed to spawn enemy!"));
		return;
	}
	spawnPosition->StarCooldown();
	float tailleCapsule = tempEnemy->GetCapsuleComponent()->GetScaledCapsuleHalfHeight();
	FVector newPosition = tempEnemy->GetActorLocation();
	tempEnemy->SetActorLocation(FVector(newPosition.X,newPosition.Y,newPosition.Z + tailleCapsule));
	
	//ajoute l'ennemis à la liste des ennemis vivants
	EnemyAliveList.Add(tempEnemy);
	
	//bind la method ennemydead à la death de l'ennemis
	tempEnemy->OnEnemyDeath.AddDynamic(this, &ALMWaveManager::EnemyDead);
	UE_LOG(LogTemp, Error, TEXT("Succes to spawn and bind on death method"));
	
	EnemySpawned++;
	EnemiesPerType[RandomIndex]--;
}

void ALMWaveManager::EnemyDead(ALMEnemy* enemyToRemoveFromLife)
{
	//remove l'ennemis de la liste des ennemis vivants
	enemyToRemoveFromLife->OnEnemyDeath.RemoveDynamic(this,&ALMWaveManager::EnemyDead);
	EnemyAliveList.Remove(enemyToRemoveFromLife);
}

bool ALMWaveManager::RemainsEnemyToSpawn()
{
	int TotalEnemies = 0;

	for (int EnemyCount : EnemiesPerType)
	{
		TotalEnemies += EnemyCount;
	}
	bool stillToSpawn = TotalEnemies !=0;
	bool no = EnemyAliveList.Num() < MaxEnemyInstantiate;
	return EnemyNumberInWave != EnemySpawned && no && stillToSpawn;
}

void ALMWaveManager::CheckForSpawnerOK()
{
	SpawnPositionsOK.Empty();
	for (ALMSpawnPosition* spawnPosition : SpawnPositionsList)
	{
		if(spawnPosition->CanSpawn())
		{
			SpawnPositionsOK.Add(spawnPosition);
		}
	}
}

void ALMWaveManager::GetRandomDataWaveRow()
{
	TArray<FName> RowNames = WaveDatePreset->GetRowNames();
	int RandomIndex = FMath::RandRange(0, RowNames.Num() - 1);;

	FName WaveName = RowNames[RandomIndex];

	CurrentWaveName = WaveName.ToString();
	FLMWaveStructure* InfoWaveRow = WaveDatePreset->FindRow<FLMWaveStructure>(WaveName, "");

	//InternalEnemyRatio = InfoWaveRow->WaveComposition;
	EnemyNumberInWave = InfoWaveRow->EnemyCount;
	MaxEnemyInstantiate = InfoWaveRow->MaxEnemyCount;
	
	for (const FLMEnemyRatio& EnemyRatio : InfoWaveRow->WaveComposition)
	{
		// Ajoute chaque EnemyType à la liste
		AllEnemyType.Add(EnemyRatio.EnemyType);
	}

	// Récupère les ratios (NbEnemy) et calcule leur somme
	int32 TotalRatio = 0;
	for (const FLMEnemyRatio& EnemyRatio : InfoWaveRow->WaveComposition)
	{
		TotalRatio += EnemyRatio.NbEnemy;
	}

	// Sécurité : Si le total est 0, évite la division par zéro
	if (TotalRatio == 0) return;

	// Répartit les ennemis proportionnellement
	int32 RemainingEnemies = EnemyNumberInWave;
	for (const FLMEnemyRatio& EnemyRatio : InfoWaveRow->WaveComposition)
	{
		// Calcul proportionnel arrondi
		int32 Count = FMath::RoundToInt((float(EnemyRatio.NbEnemy) / TotalRatio) * EnemyNumberInWave);
        
		EnemiesPerType.Add(Count);
		RemainingEnemies -= Count;
	}

	// Ajuste les arrondis pour respecter exactement EnemyNumberInWave
	for (int32 i = 0; i < RemainingEnemies; ++i)
	{
		EnemiesPerType[i % EnemiesPerType.Num()]++;
	}
}

// Called every frame
void ALMWaveManager::Tick(float DeltaTime)
{
	Super::Tick(DeltaTime);
	if(RemainsEnemyToSpawn())
	{
		CheckForSpawnerOK();
		if(SpawnPositionsOK.Num() > 0)
		{
			//Choose spawnpoint to spawn
			int32 RandomIndex = FMath::RandRange(0, SpawnPositionsOK.Num() - 1);
			SpawnEnemy(SpawnPositionsOK[RandomIndex]);
			EnemySpawned++;
		}
	}
	else
	{
		//lauch break time
		// Lance un timer de 3 secondes avant d'appeler StartWave()
		GetWorld()->GetTimerManager().SetTimer(BreakTimer, this, &ALMWaveManager::StartWave, BreakTime, false);
	}
}


void ALMWaveManager::StartWave()
{
	GetRandomDataWaveRow();
	EnemySpawned = 0;
}
